"""
@file   : color.py
@author : jiangmenggui@hosonsoft.com
@data   : 2024/4/2
"""


class _Color:

    def __init__(self, s, start, end):
        self.raw = s
        self.s = start + s + end
        self.start = start
        self.end = end

    def __len__(self):
        return self.raw.__len__()

    def __str__(self):
        return self.s.__str__()

    def __repr__(self):
        return self.s.__repr__()

    def __iter__(self):
        return self.raw.__iter__()

    def __add__(self, other):
        return self.s + other

    def __format__(self, format_spec):
        return self.start + self.raw.__format__(format_spec) + self.end


class _ColorFactory:

    def __init__(self, name, start, end='\033[0m'):
        self.name = name
        self.start = start
        self.end = end

    def __call__(self, s: str):
        return _Color(s, self.start, self.end)


class Color:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    red = _ColorFactory('red', FAIL)
    green = _ColorFactory('green', OKGREEN)
    yellow = _ColorFactory('yellow', WARNING)
    blue = _ColorFactory('blue', OKBLUE)
    cyan = _ColorFactory('cyan', OKCYAN)


if __name__ == '__main__':
    pass



# coding=utf-8
import datetime
import threading
import time
from functools import wraps
from logging import Logger
from typing import Callable

from my_tools.log_manager import LogManager


def time_cache(cache_time: float):
    __cache = {}

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            key = (func.__name__, *args, *tuple(kwargs.items()))
            if key in __cache and time.time() - __cache[key][0] <= cache_time:
                return __cache[key][1]
            res = func(*args, **kwargs)
            __cache[key] = (time.time(), res)
            return res

        return wrapper

    return decorator


_timer_logger = LogManager('timer').get_logger()
_catch_exception_logger = LogManager('catch_exception', log_file=None, error_log_file=None).get_logger()
_debug_logger = LogManager('debug').get_logger()


def timer(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        _timer_logger.info(f'{func.__name__} start')
        start = time.perf_counter()
        res = func(*args, **kwargs)
        end = time.perf_counter()
        _timer_logger.info(f'{func.__name__} end. use time {(end - start) * 1000:.4f} ms')
        return res

    return wrapper


def auto_retry(retry=3, interval=1, logger=_catch_exception_logger):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for _ in range(retry):
                try:
                    return func(*args, **kwargs)
                except Exception as e:  # noqa
                    logger.error(f'[retry]{func.__name__} error: [{e.__class__.__name__}]{str(e)}')
                    time.sleep(interval)
            return func(*args, **kwargs)

        return wrapper

    return decorator


def catch_exception(func=None, /, retry=0, interval=0):
    if func is None:
        return catch_exception(_catch_exception_logger, retry, interval)
    if isinstance(func, Logger):
        def outer(func1):
            if retry:
                func1 = auto_retry(retry, interval, func)(func1)

            @wraps(func1)
            def inner(*args, **kwargs):
                try:
                    return func1(*args, **kwargs)
                except Exception as e:
                    func.exception(f'{func1.__name__} error: [{e.__class__.__name__}]{str(e)}')

            return inner

        return outer
    elif isinstance(func, Callable):
        if retry:
            func = auto_retry(retry, interval, _catch_exception_logger)(func)

        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                _catch_exception_logger.exception(f'{func.__name__} error: [{e.__class__.__name__}]{str(e)}')

        return wrapper
    else:
        raise TypeError("catch_exception的第一个参数应该是函数或日志对象")


def debug(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        _f_info = func.__name__
        args_str = (*args, *map(lambda x: f"{x[0]}={x[1]}", kwargs.items()))
        _f_info = f'{func.__name__}({", ".join(map(str, args_str))})'
        now = datetime.datetime.now()
        t = time.perf_counter()
        res = func(*args, **kwargs)
        _debug_logger.debug(f'function execute result: \n'
                            f'  execute :  {_f_info}\n'
                            f'   result :  {str(res)[:500]}\n'
                            f' start at :  {now}\n'
                            f'finish at :  {datetime.datetime.now()}\n'
                            f' use time :  {(time.perf_counter() - t) * 1000:.4f} ms')
        return res

    return wrapper


def singleton(cls: str | type = ''):
    _instance = {}
    lock = threading.Lock()
    if isinstance(cls, str):
        def outer(cls_):
            @wraps(cls_)
            def inner(*args, **kwargs):
                # key = (cls_, getattr(cls_, cls, None))
                with lock:
                    key = kwargs.get(cls, (str(args[0]) if args else None))
                    key = (cls_, key)
                    if key not in _instance:
                        _instance[key] = cls_(*args, **kwargs)
                    return _instance[key]

            return inner

        return outer
    else:

        @wraps(cls)
        def wrapper(*args, **kwargs):
            with lock:
                if cls not in _instance:
                    _instance[cls] = cls(*args, **kwargs)
                return _instance[cls]

        return wrapper


if __name__ == '__main__':
    from my_tools.monkey import patch_print

    patch_print()

    @singleton
    class Test:

        def __init__(self, uri: str):
            pass

    t1 = Test('123')
    t2 = Test('1234')
    print(t1 is t2)




    @debug
    def fib(n):
        # if random.random() < 0.5:
        #     raise ValueError('test exception')
        s = 0
        for i in range(n):
            s += i
        return s


    @debug
    def add(a, b):
        return a + b


    fib(n=1000000)
    # time.sleep(4)
    # fib(1000000)
    print('Hello World')
    add(a=100, b=200)



# coding=utf-8
import logging
import re
import sys

from my_tools.log_manager import LogManager

_print_logger = LogManager(
    'print',
    level=logging.DEBUG,
    console_colors={"INFO": "cyan"},
    formatter='%(asctime)s.%(msecs)03d - %(name)s - %(message)s',
    log_file='/pythonlog/print.log', error_log_file=''
).get_logger()


_print_error_logger = LogManager(
    'print-error',
    level=logging.DEBUG,
    console_colors={"INFO": "cyan"},
    formatter='%(message)s',
    log_file='/pythonlog/print.log', error_log_file=''
).get_logger()


print_raw = print


def my_print(*args, sep=' ', end='\n', file=None, flush=False):
    fra = sys._getframe(1)
    line = fra.f_lineno
    file_name = fra.f_code.co_filename
    if file in [sys.stdout, None]:
        args = (f'"{file_name}:{line}" - {fra.f_code.co_name} :', *args)
        _print_logger.info(sep.join(map(str, args)))
    else:
        _print_error_logger.error(sep.join(map(str, args)).rstrip('\n'))


def patch_print():

    try:
        __builtins__.print = my_print
    except AttributeError:
        __builtins__['print'] = my_print


def restore_print():
    try:
        __builtins__.print = print_raw
    except AttributeError:
        __builtins__['print'] = print_raw


if __name__ == '__main__':
    from my_tools.decorators import catch_exception, auto_retry

    patch_print()


    @catch_exception(3)
    def func():
        print('hello world')
        raise ValueError("123")


    def to():
        func()


    to()
    print('END')
